// USB specific routines of of the CForth  project **************************
//
// Author and (c)2020-24 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.c", "main.h", "cforth.c",
// "cforth.h", "dmx.c" and "globaldefines.h"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH/Rowley Associates Ltd
//
// Contact: <www.midi-and-more.>
// A detailled hardware construction and operation manual
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32g431xx.h"
#include "globaldefines.h"

#define EP0_MAXLEN  64
#define BUSPOWERED
#define PRODSTRINGLEN  24
#define MANUSTRINGLEN  18

const uint8_t CDC_DEVICEDESCRIPTOR [0x12] =
{
0x12,                   //length of this descriptor=18 bytes
1,                      //descriptor type 	DEVICE Descriptor = 1
//0x00,0x02,                 //bdcUSB current rev of USB spec= 0x0200 = USB2.00
//(host asks for HighSpeed mode descriptor, request not supported here.)
//Therefore more easily take -- result is the same together with USB2 host:
0x10,0x01,              //bcdUSB current rev of USB spec= 0x0110 = USB1.10
2,                      //bDeviceClass	COMMUNICATION DEVICE CLASS = 2
//0,                      //bDeviceClass		Device defined at Interface level
0,       		        //bDeviceSubClass 	= 0x00 unused
0,                      //bDevice Protocol 	= 0x00, unused
EP0_MAXLEN,             //bMaxPacketSize0 for endpoint 0
//if your device shall appear as "STMicroelectronics Virtual Com Port"
DEFAULT_VIDL,DEFAULT_VIDH, //idVendor       DEFAULT_VID= 0x0483
DEFAULT_PIDL,DEFAULT_PIDH, //idProduct 	    DEFAULT_PID= 0x5740
REVNUM,HARDWARE_VERSION, //bcd_Device	release code
0,                      //iManufacturer index of manufacturer string descriptor (unused)
0,   		            //iProduct	    index of product name string descriptor = 1
0,                      //iSerialNumber	serial number (unused)
1     	                //bNumConfigurations = 1 only one available
};


const uint8_t CDC_CONFIGDESCRIPTOR [67] =
{
9,                      //bLength		    length of this descriptor = 9 bytes
2,          	        //bDescriptor type	CONFIGURATION descriptor = 2
67,0,        			//wTotalLength	    total length of configuration block=67bytes
2,                      //bNumInterfaces    two interfaces incl this descriptor
1,      		        //bConfigurationValue ID of this configuration = 1
0,                      //iConfiguration	index of string descriptor:unused=0
#ifdef BUSPOWERED
0x80,                   //bmAttributes		BUS-powered - no remote wakeup
49,         	           //MaxPower			entry*2 = 98mA
#else  //default
0xC0,                   //bmAttributes		SELF-powered - no remote wakeup
0,                       //MaxPower 			= 0 (self-powered only)
#endif

// CDC_INTERFACE_DESCRIPTOR
9,                      //bLength       	length of this descriptor = 9 bytes
4,                      //bDescriptorType	INTERFACE descriptor type=4
0,          	        //bInterfaceNumber	index of this interface =0
0,                      //bAlternateSetting	index of this setting =0
1,		                //bNumEndpoints 	(Notification endpoint) = 1
2,                      //bInterfaceClass 	= 2 (COMMUNICATION)
2,       		        //bInterfaceSubClass = 2 (ABSTRACT_CONTROL_MODEL)
0,                      //bInterfaceProtocol = 0 (unused)
0,        	            //iInterface 		= 0 (unused)

//COMMUNICATION DEVICE specific functional descriptor
5,                      //bLength			length of this descriptor = 5 bytes
0x24,       	        //bDescriptorType	CS_INTERFACEdescriptor type = 0x24
0,                      //bDescriptortSubtype HEADER subtype
0x10,1,                 //bcdADC	=0x0100 revision of class specific descriptors

//CALL MANAGEMENT functional Descriptor
5,                      //bLength           length of this descriptor = 5 bytes
0x24,                   //bDescriptorType	CS_INTERFACEdescriptor type = 0x24
1, 			            //bDescriptorSubtype
3,                      //bmCapabilities	Device handles call management by itself
                        //ia DataInterface
						//i.e. in this case: no call management required
0,                     	//bDataInterface:	Data interface no.0 (see below) associated
//1,                     	//bDataInterface:	Data interface no.0 (see below) associated

//ABSTRACT CONTROL management functional Descriptor
4,                      //bLength			length of this descriptor = 4 bytes
0x24,         	        //bDesriptorType	CS_INTERFACEdescriptor type = 0x24
2,                      //bDescriptorSubtype
7,				        //bmCapabilities    Send_Break,Set_Line_Coding... etc

//UNION functional descriptor
5,                      //bLength			length of this descriptor = 5 bytes
0x24,                   //bDescriptorType	CS_INTERFACE descriptor
6,                      //bDescriptorSubtype UNION FUNCTIONAL descriptor
0,		       	        //bMasterInterface
1,                      //bSlaveInterface0

//Notification standard endpoint descriptor
7,  			        //bLength           length of this descriptor = 7 bytes
5,                      //bDescriptorType	ENDPOINT
0x83,			        //bEndpointAddress	no.3, IN
3,                      //bmAttributes 	INTERRUPT = 3
16,0,                    //wMaxPacketSize	=16
64,			            //interval in ms    =64

//DATA CLASS INTERFACE descriptor
9,                      //bLength			length of this descriptor = 9 bytes
4,				        //bDescriptorType	INTERFACE descriptor type
1,                      //bInterfaceNumber
0,				        //bAlternateSetting
2,                      //bNumEndpoints
0x0A,       			//bInterfaceClass	Data Class
0,                      //bInterfaceSubclass unused
0,				        //bInterfaceProtocol no specific protocol
0,                      //iInterface 		unused

//DATA BULK IN endpoint standard descriptor
7,				        //bLength           length of this descriptor = 7 bytes
5,                      //bDescriptorType	ENDPOINT
0x81,			        //bEndpointAddress	nr.1, IN
2,                      //bmAttributes		=2 (bulk)
0x40,0,			        //wMaxPacketSize	=64
0,				        //bInterval			=0:ignore for bulk transfer (vorher:1ms)

//DATA BULK OUT endpoint standard descriptor
7,                      //bLength			length of this descriptor = 7 bytes
5,				        //bDescriptorType	ENDPOINT
2,                      //bEndpointAddress	nr.=2, OUT
2,				        //bmAttributes 		=2 (bulk)
0x40,0,			        //wMaxPacketSize	=64
0				        //bInterval			=0:ignore for bulk transfer (vorher:1ms)
};

//Manufacturer String Descriptor index 2
const uint8_t MANUFACT_STRINGDESCRIPTOR [MANUSTRINGLEN] =
{
MANUSTRINGLEN,			//bLength = 16+2  Descriptor length in Bytes
0x03,                   //bDescriptorType = 0x03, Constant String = 3
'h',0,'o',0,'m',0,'e',0,'m',0,'a',0,'d',0,'e',0
};

//Product String Descriptor index 1
const uint8_t PRODUCT_STRINGDESCRIPTOR [PRODSTRINGLEN] =
{
PRODSTRINGLEN,			//bLength = 22+2  Descriptor length in Bytes
0x03,                   //bDescriptorType = 0x03, Constant String = 3
'C',0,'F',0,'o',0,'r',0,'t',0,'h',0,'-',0,'M',0,'I',0,'D',0,'I',0
                        //bString (Unicode 16Bit)
};

const uint8_t LINE_CODING_115200 [7] =
{
//0x00,0x00,0x00,0x00,0,0,0
0x00,0xC2,0x01,0x00,0,0,8  //0x0001C200=decimal 115200, rest describes 1N8
//the baud rate is faked, the microcontroller accepts any and sets its own for the USART.
//but 115200 seems to be the best default at host side.
//Else enter your baudrate 32bit little endian formatted.
};

//Device Status: bus powered, no remote wakeup
const uint8_t DEVICE_STATUS[2] = 
{
0,0						//Bus powered, no Remote Wakep
};

//asked by "Raspbian"
const uint8_t COUNTRY[4] = 
{
4,3,9,4						//inserts country code 0409 in other string req
};

const uint8_t MIDI_DEVICEDESCRIPTOR[0x12] = //dez18
{
18,        	            //bLength			length of this descriptor=18 bytes
1,                     	//descriptor type 	DEVICE Descriptor = 1
//0x00,0x02,                 //bdcUSB current rev of USB spec= 0x0200 = USB2.00
0x10,0x01,              //bcdUSB current rev of USB spec= 0x0110 = USB1.10
0,      		    //bDeviceClass		Device defined at Interface level
0,                     	//bDeviceSubClass 	= 0x00 unused
0,	    	            //bDevice Protocol 	= 0x00, unused
EP0_MAXLEN,	            //bMaxPacketSize0	Endp.0 = 64
DEFAULT_VIDL,DEFAULT_VIDH, //idVendor       DEFAULT_VID= 0x0483
DEFAULT_PIDL+MIDI_PLUS,DEFAULT_PIDH, //idProduct DEFAULT_PID=0x5743! else host problem!
REVNUM,HARDWARE_VERSION, //bcd_Device		release code
2,   		            //Manufacturer		index of manufacturer string descriptor (unused)
1,                   	//iProduct			Index of product name string descriptor (=1)
0,     		            //iSerialNumber		serial number (unused)
1                    	//bNumConfigurations = 1 only one available
};


const uint8_t MIDI_CONFIGDESCRIPTOR[0x65] =  //dez101
{
9,                      //bLength			length of this descriptor = 9 bytes
2,                      //bDescriptorType	CONFIGURATION descriptor type = 2
0x65,0,			        //wTotalLength		length of configuration block =101
                        //					incl this descriptor
2,      		        //bNumInterfaces    two interfaces
1,						//bConfigurationValue ID of this configuration
0,                      //iConfiguration	index of string descriptor:unused=0
#ifdef BUSPOWERED
0x80,                    //bmAttributes		BUS-powered - no remote wakeup
49,                      //MaxPower 		entry*2 = 98mA
#else
0xC0,                   //bmAttributes		SELF-powered - no remote wakeup
0,                      //MaxPower			= 0 (self-powered only)
#endif
//Standard Audio Class Interface Descriptor
9,                      //bLength			length of this descriptor = 9 bytes
4,          	        //bDescriptorType	INTERFACE descriptor type=4
0,                      //bInterfaceNumber	index of this interface
0,	                    //bAlternativeSetting index of this setting
0,	                    //bNumEndpoints		= 0 (EP0 only)
1,       		        //bInterfaceClass	= 0x01 (AUDIO)
1,                      //bInterfaceSubClass = 0x01 (AUDIO_CONTROL)
0,         	            //bInterfaceProtocol = 0 (unused)
0,                      //iInterface 		= 0 (unused)
//Class-specific AudioControl Interface descriptor
9,       	            //bLength			length of this descriptor = 9 bytes
0x24,                   //bDescriptorTye	CS_INTERFACEdescriptor type = 0x24
1,       		        //bDescriptorSubtyte HEADER subtype
0,1,                    //bcdADC			=0x0100 revision of class specific descriptors
9,0,                    //wTotalLength		= 9
1, 			            //bInCollection		number of streaming interfaces
1,   			        //baInterfaceNr(1)	MIDIStreaming interface no.1
						//					belongs to this Audio Control
//Standard MIDI Streaming Interface Descriptor
9,                      //bLength	        length of this descriptor = 9 bytes
4,         	            //bDescriptorType	INTERFACE descriptor type = 0x04
1,                      //ibInterfaceNumber	index of this interface
0,				        //bAlternativeSetting index of this setting
2,						//bNumENdpoints		2 endpoints
1, 			            //bInterfaceClass	AUDIO
3, 						//bInterfaceSubclass MIDISTREAMING
0,          	        //bInterfaceProtocol = 0 (unused)
0,                       //iInterface 		= 0 (unused)
//Class specific MIDI Streaming Interface Descriptor
7,                      //bLength 			length of this descriptor = 7 bytes
0x24,                   //bDescriptorType	CD_INTERFACE descriptor
1,       	            //bDescriptorSubtye	MS_HEADER subtype
0,0x01,                 //bcdADC			Revision of this class specification 0x0100
0x41,0x00,		        //wTotalLength		total size of class-specific descriptors
//MIDI IN Jack descriptor (embedded)
6,                      //bLength           length of this descriptor = 6 bytes
0x24,			        //bDescriptorType	CS_INTERFACE descriptor
2,						//bDescriptorSubtype MIDI_IN_JACK subtype
1,				        //bJackType			EMBEDDED
1,						//bJackID			ID of this jack
0,				        //iJack 			unused
//MIDI IN Jack descriptor (external)
6,			            //bLength           length of this descriptor = 6 bytes
0x24,			        //bDescriptorType	CS_INTERFACE descriptor
2,						//bDescriptorSubtype MIDI_IN_JACK subtype
2,				        //bJackType			EXTERNAL
2, 						//bJackID			ID of this jack
0,				        //iJack 			unused
//MIDI OUT Jack descriptor (embedded)
9,			            //bLength           length of this descriptor = 9 bytes
0x24,			        //bDescriptorType	CS_INTERFACE descriptor
3, 						//bDescriptorSubtype MIDI_OUT_JACK subtype
1,				        //bJackType			EMBEDDED
3, 						//bJackID			ID of this jack
1,				        //bNrInputPins		= 1
2,						//BaSourceID(1)		this jack is connected to jack with ID 2
1,				        //BaSourcePin(1)	there connected to pin 1
0,   					//iJack 			unused
//MIDI OUT Jack descriptor (external)
9,			            //length of this descriptor = 9 bytes
0x24,					//bDescriptorType	CS_INTERFACE descriptor
3,				        //bDescriptorSubtype MIDI_OUT_JACK subtype
2, 						//bJackType			EXTERNAL
4,				        //bJackID			ID of this jack
1,						//bNrInputPins		= 1
1,				        //BaSourceID(1)		this jack is connected to jack with ID 1
1,						//BaSourcePin(1)	there connected to pin 1
0,  			        //iJack 			unused
//Standard BULK IN endpoint descriptor
9,						//bLength           length of this descriptor = 9 bytes
5,			            //bDescriptorType	ENDPOINT descriptor
0x81,					//bEndpointAddress	IN endpoint no.1
2,			            //bmAttributes		BULK DATA, not shared
0x40,0x00,				//wMaxPacketSize	64 bytes/packet
0,			            //bInterval 		ignored for Bulk = 0
0,				        //bRefresh 			unused
0,						//bSynchAddress 	unused
//Class-specific BULK IN endpoint descriptor
5,			            //bLength			length of this descriptor = 5 bytes
0x25,					//bDescriptorType	CS_ENDPOINT descriptor
1,				        //bDescriptorSubtype MS_GENERAL subtype
1,						//bNumEmbMIDIJack	= 1 number of embedded MIDI OUT jacks
3,				        //BaAssocJackID(1)	associated with embedded MIDI OUT no.3
//Standard BULK OUT endpoint descriptor
9,						//bLength           length of this descriptor = 9 bytes
5,				        //bDescriptorType	ENDPOINT descriptor
2,   					//bEndpointAddress	OUT endpoint no.2
2,			            //bmAttributes		BULK DATA,not shared
0x40,0x00,				//wMaxPacketSize	64 bytes/packet
0,			            //bInterval 		ignored for Bulk = 0
0,				        //bRefresh 			unused
0,						//bSynchAddress 	unused
//Class-specific BULK OUT endpoint descriptor
5,                      //bLength			length of this descriptor = 5 bytes
0x25,			        //bDescriptorType	CS_ENDPOINT descriptor
1,				        //bDescriptorSubtype MS_GENERAL subtype
1,						//bNumEmbMIDIJack	= 1 number of embedded MIDI IN jacks
1				        //baAssocJackID(1)	associated with embedded MIDI IN no.1
};

//abbreviations recalculated from stm32g431xx.h
#define CLEAR_CTR_RX_MASK  (USB_EPREG_MASK & (~USB_EP_CTR_RX))  //=0x0F8F
#define CLEAR_CTR_TX_MASK  (USB_EPREG_MASK & (~USB_EP_CTR_TX))  //=0x8F0F
//assignment of buffer offset and corresponding buffer sizes
#define EP0_TX_OFFSET   0x40    //buffer size 64 bytes
#define EP0_RX_OFFSET   0x80    //buffer size 64 bytes
#define EP1_TX_OFFSET  0x100    //buffer size 128 bytes
#define EP2_RX_OFFSET  0x200    //buffer size 128 bytes
#define EP3_TX_OFFSET  0x300    //buffer size 32 bytes


// ***************************************

extern uint32_t  gVid, gPid;
#ifndef MINI
extern uint32_t gSplit;
#endif
extern uint32_t gfTermCtrlBD;	//message from terminal input for backgnd proc

extern void LED_Off(void);
extern void msDelay(uint16_t);
extern void wait5us(void);
extern void servSystem(void);
extern void sendB(uint8_t);
//extern void sendX8(uint8_t);
//extern void sendX16(uint16_t);
//extern void sendX32(uint32_t);
//extern uint32_t pullUsartRxBuf(void);

//from cforth.c
extern uint32_t gfTermCtrlBD;	//message from terminal input for backgnd proc
extern void op_DEBUG(void);

// ***************************************

//Endpoint 0
uint16_t  *gUSB_ADDR0_TX, *gUSB_COUNT0_TX, *gUSB_ADDR0_RX, *gUSB_COUNT0_RX ;
//Endpoint 1 is single buffer IN
uint16_t  *gUSB_ADDR1_TX, *gUSB_COUNT1_TX;
//Endpoint 2 is single buffer OUT
uint16_t  *gUSB_ADDR2_RX, *gUSB_COUNT2_RX;
//Endpoint 3 is single buffer IN
uint16_t  *gUSB_ADDR3_TX, *gUSB_COUNT3_TX;

#ifdef DMX
#define USB_BUFLEN 1152
#else
#define USB_BUFLEN 2044
#endif

uint8_t gUSB_TXBUF[USB_BUFLEN+4];    //for EP1 IN
volatile uint8_t *gUSB_TXHEAD=0,*gUSB_TXTAIL=0,*gUSB_TXBUFTOP=&(gUSB_TXBUF[USB_BUFLEN]);
uint8_t gUSB_RXBUF[USB_BUFLEN+4];    //for EP2 OUT
volatile uint8_t *gUSB_RXHEAD=0,*gUSB_RXTAIL=0,*gUSB_RXBUFTOP=&(gUSB_RXBUF[USB_BUFLEN]);

uint16_t * guBTABLE_ENTRY;
volatile uint32_t guAddress;

//USB Std Request parameters
volatile uint8_t   guRequestType;
volatile uint8_t   guRequest;
volatile uint16_t  guValue;
volatile uint8_t   guDescriptor;
volatile uint8_t   guDescIndex;
volatile uint16_t  guIndex;
volatile uint16_t  guLength;
// USB packet handling
volatile uint32_t   guSetupAction;   //action to be performed in servEP0_In() routine
volatile uint32_t   guTx0RemainCnt;  //remaining number of bytes to be transmitted at EP0
const uint8_t      *guInSendPtr;    //pointer to storage of next EP0 data to be transmitted
uint32_t            guDeviceClass, guConfigured, guLockEP2;
uint8_t            *guEP2Ptr;       //pointer to EP2_OUT fifo

//MIDI special
uint8_t	guMSTATUSok, guMDATA1ok, guMDATA2ok, guSYSEXBODY, guTxPackCnt;
uint8_t guMSTATUS,guMIDIDAT1, guMIDIDAT2, guPACKTYPE;
uint8_t guRx2RunState;   //running state of EP2 MIDI Messages

void stopUSB(void);
void initUSB(void);
void subInitUSB(void);

//void USB_IRQHandler(void);
void servEP0_Out(void);
void servEP0_In(void);
void prepareDescriptor(void);
void sendZLP_EP0(void);
void stall_EP0(void);
void servEP1_In(void);
void servEP2_Out(void);
void servEP3_In(void);
uint32_t pullUsbRxBuf(void);
uint8_t deCodeMidi(uint8_t);
void sendUsbByte(uint8_t);
void enCodeMidi(uint8_t);


//###########################################################################

void stopUSB(void)   // *****************************************************
//  ***************************************************************************
{
    GPIOA->AFR[0] &= 0x00FFFFFF;                //de-configure GPIOA 11,12 as AF
    GPIOA->MODER&=~(GPIO_MODER_MODER11|GPIO_MODER_MODER12); //now input
    NVIC_DisableIRQ(USB_HP_IRQn);               //ever used ??
    NVIC_DisableIRQ(USB_LP_IRQn);
    RCC->APB1RSTR1 |= RCC_APB1RSTR1_USBRST;     //reset USB
    USB->CNTR |= USB_CNTR_PDWN;                 //USB is disabled completely
    RCC->APB1ENR1 &= ~RCC_APB1ENR1_USBEN;       //disable USB clock
    guConfigured = 0;
}


void initUSB(void)  // ******************************************************
// **************************************************************************
{
uint32_t k;
uint16_t * tempPtr;

    RCC->APB1ENR1 |= RCC_APB1ENR1_USBEN;
    RCC->APB1RSTR1 |= RCC_APB1RSTR1_USBRST;
    wait5us();
    RCC->APB1RSTR1 &= ~RCC_APB1RSTR1_USBRST;

//as a precaution, init PA11,12 as high impedance inputs
    GPIOA->MODER &= ~(GPIO_MODER_MODE11 | GPIO_MODER_MODE12);
                                                // config pin 11,12 as input
//GPIOA->OSPEEDR is 2MHz by default
    GPIOA->PUPDR &=  ~(GPIO_PUPDR_PUPD11 | GPIO_PUPDR_PUPD12);

	guSetupAction = 0;

    tempPtr= (uint16_t *) USB_BASE;
    for (k=0; k<0x10; k++)  *(tempPtr+k) = 0;
    for (k=0x22; k<0x30; k++)  *(tempPtr+k) = 0;

// init endpoint registers
    USB->BCDR |= USB_BCDR_DPPU;      //enable internal pull-up at USB-DP
    USB->CNTR = USB_CNTR_FRES;       //switch USB power on but keep forced reset state
    msDelay(200);
    USB->CNTR = 0x0000;          //terminate forced reset
    USB->ISTR = 0x0000;          //remove spurious pending interrupts

    subInitUSB();
    NVIC_EnableIRQ(USB_LP_IRQn);
}


void subInitUSB(void)  // ****************************************************
// *************************************************************************
{

    USB->DADDR = USB_DADDR_EF;      // =0x80 and addr=0
    guAddress = 0;
    USB->BCDR |= USB_BCDR_DPPU;     //enable internal pull-up at USB-DP
    USB->BTABLE = 0x00;             //init USB_BTABLE register
                            //(address relative to start of dedicated packet memory)
//init "buffer description table"
    guBTABLE_ENTRY = (uint16_t *) (USB_PMAADDR + USB->BTABLE);

//because serial speed is much less than USB speed, only single buffering is used:
    gUSB_ADDR0_TX = guBTABLE_ENTRY;
    gUSB_COUNT0_TX = guBTABLE_ENTRY+1;
    gUSB_ADDR0_RX = guBTABLE_ENTRY+2;
    gUSB_COUNT0_RX = guBTABLE_ENTRY+3;
    gUSB_ADDR1_TX = guBTABLE_ENTRY+4;
    gUSB_COUNT1_TX = guBTABLE_ENTRY+5;
    gUSB_ADDR2_RX = guBTABLE_ENTRY+10;
    gUSB_COUNT2_RX = guBTABLE_ENTRY+11;
    if (guDeviceClass == CDC_CLASS)  {
        gUSB_ADDR3_TX = guBTABLE_ENTRY+12;
        gUSB_COUNT3_TX = guBTABLE_ENTRY+13;
    }
//entries for double buffers initalized formally:
    *gUSB_ADDR0_TX = EP0_TX_OFFSET; //corresponds with ref.manual p.1967:
                                    //ADDR0_TX, buffer size 64 bytes
    *gUSB_COUNT0_TX = 0;            //COUNT0_TX
    *gUSB_ADDR0_RX = EP0_RX_OFFSET; //ADDR0_RX, buffer size 64 bytes
    *gUSB_COUNT0_RX = 0x8400;       //COUNT0_RX - binary 1 000 0100  0000 0000
    *gUSB_ADDR1_TX = EP1_TX_OFFSET; //ADDR1_TX_0, buffer size 128 bytes, IN endpoint
    *gUSB_COUNT1_TX = 0;             //COUNT1_TX_0
    *(guBTABLE_ENTRY+6) = EP1_TX_OFFSET; //ADDR1_TX_1, buffer size 128 bytes
    *(guBTABLE_ENTRY+7) = 0;              //COUNT1_TX_1
    *(guBTABLE_ENTRY+8) = EP2_RX_OFFSET; //ADDR2_2RX_1, buffer size 128 bytes
    *(guBTABLE_ENTRY+9) = 0;         //COUNT2_RX_
    *gUSB_ADDR2_RX = EP2_RX_OFFSET; //ADDR2_RX_0, buffer size 64 bytes, OUT endpoint
    *gUSB_COUNT2_RX = 0x8400;        //COUNT2_RX_0
    if (guDeviceClass == CDC_CLASS)  {
        *gUSB_ADDR3_TX = EP3_TX_OFFSET; //ADDR3_TX_0, buffer size 32 bytes, IN endpoint
        *gUSB_COUNT3_TX = 0;             //COUNT3_TX_0
        *(guBTABLE_ENTRY+14) = EP3_TX_OFFSET; //ADDR3_TX_1, buffer size 32 bytes
        *(guBTABLE_ENTRY+15) = 0;        //COUNT3_TX_1
    }
// init endpoint registers
    USB->EP0R =  0x3220;        // Control, everything prepared for first Setup token
    USB->EP1R =  0x3031;        // Bulk IN
    USB->EP2R =  0x3032;        // Bulk OUT
    if (guDeviceClass == CDC_CLASS)  USB->EP3R =  0x3633;    // Interrupt IN

    USB->CNTR = USB_CNTR_CTRM | USB_CNTR_ERRM | USB_CNTR_WKUPM | USB_CNTR_SUSPM | USB_CNTR_RESETM \
            | USB_CNTR_SOFM  | USB_CNTR_ESOFM;  //set appropriate interrupt mask

    guTx0RemainCnt = 0;

    guConfigured = 0;
    gUSB_TXHEAD = gUSB_TXBUF;   //flush EP1 application level buffer
    gUSB_TXTAIL = gUSB_TXBUF;
    gUSB_RXHEAD = gUSB_RXBUF;   //flush EP2 application level buffer
    gUSB_RXTAIL = gUSB_RXBUF;
    guTxPackCnt = 0;            //prepare enCodeMidi()
    guRx2RunState = 0;          //prepare deCodeMidi()
}

//##########################################################################

//the reset is done by Interrupt Handling
//***************************************

void USB_LP_IRQHandler(void)   // ***********************************************
// ***************************************************************************
{
volatile uint32_t  EPindex, ioDir, wIstr;

	wIstr = (uint32_t)USB->ISTR;
    if ((wIstr & USB_ISTR_CTR) != 0)  {          //if "Correct Transfer"
		EPindex = wIstr & 0x0000000F;
        ioDir = wIstr & 0x00000010;

//check project specific endpoints:

        if (EPindex == 1)  servEP1_In();
// **********************

        else if (EPindex == 2)  servEP2_Out();
 // **************************

        else if (EPindex == 0 ) {
// *****************************
 //check if SETUP and/or CTR_RX are set
            if (ioDir == 0)  {   //ioDir==0, i.e. IN packet
                servEP0_In();
            }
            else servEP0_Out();
        }

        else if ((EPindex == 3) && (guDeviceClass == CDC_CLASS))  servEP3_In();
// ****************************

        else {;}        //ignore
// *****************************
    }
    else if (USB->ISTR & USB_ISTR_RESET)  {
        subInitUSB();
        USB->ISTR = ~USB_ISTR_RESET;
    }
    else if (USB->ISTR & USB_ISTR_SUSP)  USB->ISTR = ~USB_ISTR_SUSP; 
    else USB->ISTR = 0x807F;                     //reset all 'rc_w0' type flags
}


void servEP0_Out(void)   // **************************************************
// ***************************************************************************
{
uint32_t len, k;
volatile uint16_t regVal;
uint8_t tempConfigBuf[0x12];

//check if SETUP is set
    if (USB->EP0R & USB_EP_SETUP)  {
//SETUP packet
        guRequestType = *((uint8_t*)(USB_PMAADDR + EP0_RX_OFFSET));
        guRequest = *((uint8_t*)(USB_PMAADDR + EP0_RX_OFFSET +1));
        guValue = *((uint16_t*)(USB_PMAADDR + EP0_RX_OFFSET +2));
        guDescriptor = (uint8_t)((guValue >> 8) & 0x00FF);
        guDescIndex = (uint8_t)(guValue & 0x00FF);
        guIndex = *((uint16_t*)(USB_PMAADDR + EP0_RX_OFFSET +4));
        guLength = *((uint16_t*)(USB_PMAADDR + EP0_RX_OFFSET +6));
        guSetupAction = 0;
        *gUSB_COUNT0_RX = 0x8400;                       //neutralize

        if ((guRequestType == 0x21) && (guDeviceClass == CDC_CLASS)) {
			if (guRequest == 0x20)  { 		// special case SET_LINE_CODING 
//make EP RX "VALID":
				regVal = USB->EP0R;
				if ((regVal & USB_EPRX_DTOG2)==0) USB->EP0R = 0xA280; 
													//USB_EPRX_DTOG2=0x2000
				wait5us();
				regVal = USB->EP0R;
				if ((regVal & USB_EPRX_DTOG1)==0) USB->EP0R = 0x9280; 
													//USB_EPRX_DTOG2=0x1000
				wait5us();
				regVal = USB->EP0R;
				if (regVal & USB_EP_DTOG_RX) USB->EP0R = 0x0280;
				wait5us();
				while (USB->EP0R & 0x8000)  {
					wait5us();
				}
				wait5us();
                return;

            }
            else if (guRequest == 0x22)  {    	// special case SET_LINE_STATE
				USB->EP0R = USB->EP0R & CLEAR_CTR_RX_MASK;	//clear CTR_RX		
				sendZLP_EP0();
			}
            else {
                stall_EP0();
                return;
            }
		}

        else if ((guRequestType == 0xA1) && (guRequest == 0x21) \
											&& (guDeviceClass == CDC_CLASS))  {
												//special case GET_LINE_CODING
			USB->EP0R = USB->EP0R & CLEAR_CTR_RX_MASK;      //clear CTR_RX
            guTx0RemainCnt = 7;
            if (guLength < guTx0RemainCnt) guTx0RemainCnt = guLength;
            guInSendPtr = LINE_CODING_115200;
            prepareDescriptor();
        }
        else if (((guRequestType & 0x61) == 0) || ((guRequestType & 0x61) == 1))  {
			USB->EP0R = USB->EP0R & CLEAR_CTR_RX_MASK;      //clear CTR_RX
            switch(guRequest)  {
            case 0:             //Get_Status
				guTx0RemainCnt = 2;
				guInSendPtr = DEVICE_STATUS;
				prepareDescriptor();
                break;
            case 0x05:          //Set_Address
                guTx0RemainCnt = 0;
                sendZLP_EP0();
                guAddress = guValue & 0x7F;          //keep new address globally
                guSetupAction = 5;
                break;
            case 0x06:           //Get_Descriptor
//                guSetupAction = 6;
                if (guDescriptor == 1)  {
              		guTx0RemainCnt = 0x12;       //18
					if (guDeviceClass == MIDI_CLASS) {
                        guInSendPtr = MIDI_DEVICEDESCRIPTOR;
                    }
                    else  guInSendPtr = CDC_DEVICEDESCRIPTOR;

 //if configured individually, assemble user defined Vid/Pid
                    if ((gVid != DEFAULT_VID) || (gPid != DEFAULT_PID)) {
                                 //Vid/Pid in flash  is modified by user
						for (k=0;k<0x12;k++)  {
							tempConfigBuf[k] = guInSendPtr[k];
						}
                        tempConfigBuf[8] =
                                (uint8_t)(gVid & 0x000000FF);
                        tempConfigBuf[9] =
                                (uint8_t)((gVid>>8) & 0x000000FF);
                       
						if (guInSendPtr == MIDI_DEVICEDESCRIPTOR)  {
							tempConfigBuf[10] =
								(uint8_t)((gPid & 0x000000FF) + MIDI_PLUS);
						}
						else  {
							tempConfigBuf[10] =
										(uint8_t)(gPid & 0x000000FF);
						}
						tempConfigBuf[11] =
                                (uint8_t)((gPid>>8) & 0x000000FF);

						guInSendPtr = tempConfigBuf;
					}
                }
                else if (guDescriptor == 2)  {

                    if (guDeviceClass == MIDI_CLASS) {
                        guTx0RemainCnt = 0x65;       //101
                        guInSendPtr = MIDI_CONFIGDESCRIPTOR;
                    }
                    else  {
                        guTx0RemainCnt = 0x43;       //67
                        guInSendPtr = CDC_CONFIGDESCRIPTOR;
                    }
                    if (guLength < guTx0RemainCnt) guTx0RemainCnt = guLength;
                }
                else if (guDescriptor == 3)  {
                    if (guDescIndex == 1)  {
						guTx0RemainCnt = PRODSTRINGLEN;
						if (guLength <guTx0RemainCnt) guTx0RemainCnt=guLength;
						guInSendPtr = PRODUCT_STRINGDESCRIPTOR;
					}
                    else if (guDescIndex == 2)  {
						guTx0RemainCnt = MANUSTRINGLEN;
						if (guLength <guTx0RemainCnt) guTx0RemainCnt=guLength;
						guInSendPtr = MANUFACT_STRINGDESCRIPTOR;
					}
                    else if (guDescIndex == 0)  {
						guTx0RemainCnt =4;
						if (guLength <guTx0RemainCnt) guTx0RemainCnt=guLength;
						guInSendPtr = COUNTRY;
					}
					else  {
						stall_EP0();
						return;
					}
                }
                else  {
                    stall_EP0();
                    return;
                }
                prepareDescriptor();
                break;
            case 0x09:          //Set_Configuration
                guTx0RemainCnt = 0;
                sendZLP_EP0();
                guConfigured = 0xFF;
                break;
			
            default:
				if (guRequestType == 0) {
					USB->EP0R = USB->EP0R & CLEAR_CTR_RX_MASK;     
															//clear CTR_RX	
					sendZLP_EP0();
					return;
				}
                stall_EP0();      //unsupported: stall
                break;
            }
        }
        else  {
			stall_EP0();
		}
    }

    else   {
//OUT packet, not SETUP
//*********************
//make EP RX "VALID":
		regVal = USB->EP0R;
		if (regVal & 0x8000) USB->EP0R = regVal & CLEAR_CTR_RX_MASK;
		sendZLP_EP0();
    }
}


void prepareDescriptor(void)   // ********************************************
// ***************************************************************************
{
volatile uint16_t regVal;
uint16_t k, temp1, temp2;
uint16_t * EP0InPtr;

//analyse type of request
    EP0InPtr = (uint16_t*) (USB_PMAADDR + EP0_TX_OFFSET);

    if (guTx0RemainCnt > EP0_MAXLEN) {
        *gUSB_COUNT0_TX = EP0_MAXLEN;
        guTx0RemainCnt -= EP0_MAXLEN;
    }
    else   {
       *gUSB_COUNT0_TX =guTx0RemainCnt;
        guTx0RemainCnt = 0;
    }
    for (k=0;k<((*gUSB_COUNT0_TX+1)/2);k++)  {
        temp1 = (uint16_t)*guInSendPtr;
        guInSendPtr++;
        temp2 = (uint16_t)*guInSendPtr;
        guInSendPtr++;
        *EP0InPtr = temp1 | temp2<<8;
        EP0InPtr++;
    }
//make EP0 TX "VALID":
    if (USB->EP0R & 0x0080)  USB->EP0R = USB->EP0R & CLEAR_CTR_TX_MASK; //clear CTR_TX
    *gUSB_COUNT0_RX = 0x8400;    //neutralize (see "DCD_EP_PrepareRx()")
    regVal = USB->EP0R;
    if ((regVal & USB_EPTX_DTOG2)==0) USB->EP0R = 0x82A0; //USB_EPTX_DTOG2=0x0020
    regVal = USB->EP0R;
    if ((regVal & USB_EPTX_DTOG1)==0) USB->EP0R = 0x8290; //USB_EPTX_DTOG2=0x0010
}

void sendZLP_EP0(void)   // **************************************************
// ***************************************************************************
{
volatile uint16_t regVal;

    *gUSB_COUNT0_TX = 0;

//make EP TX "VALID":
    regVal = USB->EP0R;
    if (regVal & 0x0080)  USB->EP0R = regVal & CLEAR_CTR_TX_MASK; //clear CTR_TX
//    *gUSB_COUNT0_TX = 0;
    regVal = USB->EP0R;
    if ((regVal & USB_EPTX_DTOG2)==0) USB->EP0R = 0x82A0; //USB_EPTX_DTOG2=0x0020
    regVal = USB->EP0R;
    if ((regVal & USB_EPTX_DTOG1)==0) USB->EP0R = 0x8290; //USB_EPTX_DTOG2=0x0010
//Status stage will be followed by Setup Stage
//    regVal = USB->EP0R;
//    if (regVal & USB_EP_DTOG_RX) USB->EP0R = 0xC280;      //EP_DTOG_RX=0x4000
}

void servEP2_Out(void)   // **************************************************
// ***************************************************************************
{
volatile uint16_t regVal, fillstate;
uint8_t len;

    if (USB->EP2R & 0x8000) USB->EP2R = USB->EP2R & CLEAR_CTR_RX_MASK;

    guEP2Ptr = (uint8_t *) (USB_PMAADDR + EP2_RX_OFFSET);
    len = *gUSB_COUNT2_RX & 0x03FF;

//restore endpoint buffer into cyclic application level buffer
//    while ((len > guDeviceClass*3) && (len < 65 )) {
    while ((len > (guDeviceClass+(guDeviceClass<<1))) && (len < 65 )) { //*3
                                //lower limit:CDC_CLASS:>0, MIDI_CLASS:>3
                                // limits are protection against underflow
        if (guDeviceClass == MIDI_CLASS)  {
                            //transform USB MIDI packets into MIDI byte stream
                            //regular MIDI packets always consist of 4 bytes
            len = deCodeMidi(len) ;
        }
        else  {  //CDC_CLASS
//modification 11June22
#ifdef MINI
			if (*guEP2Ptr == CTRL_B)  gfTermCtrlBD = CTRL_B;
			else if (*guEP2Ptr == CTRL_D)  gfTermCtrlBD = CTRL_D;
#else
			if ((gSplit != 2) && (*guEP2Ptr == CTRL_B))  gfTermCtrlBD = CTRL_B;
			else if ((gSplit != 2)&&(*guEP2Ptr==CTRL_D)) gfTermCtrlBD = CTRL_D;
#endif
			else  {
				*gUSB_RXHEAD++ = *guEP2Ptr;
				if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
			}
			guEP2Ptr++;
            len-- ;
        }
    }

//calculate fill level of gUSB_RXBUF
//TAIL>HEAD?
    if (gUSB_RXTAIL > gUSB_RXHEAD) {
        fillstate = (uint32_t)gUSB_RXHEAD + USB_BUFLEN - (uint32_t)gUSB_RXTAIL ;
    }
    else fillstate = (uint32_t)gUSB_RXHEAD - (uint32_t)gUSB_RXTAIL ;

// lock EP2 interrupt when EP2_BUF contains at least (SIZE-64) bytes
		// gets unlocked in routine pullUsbRxBuf
                    // as soon as EP2_BUF contains less than (SIZE-64) bytes
                    //this way EP2 always can accept
                    //another packet of 64 bytes without data loss
                    //(similar to CTS/RTS handshake at RS-232)
    guLockEP2 = 0;
    if (fillstate > USB_BUFLEN - 70)  {
        guLockEP2 = 0xFF;
        return;
    }

//make EP RX "VALID" again:
    regVal = USB->EP2R;
    if ((regVal & USB_EPRX_DTOG2)==0) USB->EP2R = 0xA082; //USB_EPRX_DTOG2=0x2000
	regVal = USB->EP2R;
    if ((regVal & USB_EPRX_DTOG1)==0) USB->EP2R = 0x9082; //USB_EPRX_DTOG2=0x1000
}


void servEP0_In(void)   // ***************************************************
// ***************************************************************************
{
volatile uint16_t regVal;

    USB->EP0R = USB->EP0R & CLEAR_CTR_TX_MASK;      //clear CTR_TX=sendACK

    if (guSetupAction == 5)  {
        USB->DADDR = USB_DADDR_EF | guAddress;       // =0x80 plus new addr
        guSetupAction = 0;
    }
    if (guTx0RemainCnt == 0) {
//make EP RX "VALID":
        if (USB->EP0R & 0x8000) USB->EP0R = USB->EP0R & CLEAR_CTR_RX_MASK;
        regVal = USB->EP0R;
        if ((regVal & USB_EPRX_DTOG2)==0) USB->EP0R = 0xA280; //USB_EPRX_DTOG2=0x2000
        regVal = USB->EP0R;
        if ((regVal & USB_EPRX_DTOG1)==0) USB->EP0R = 0x9280; //USB_EPRX_DTOG2=0x1000
		*gUSB_COUNT0_TX = 0;
    }
    else prepareDescriptor();   //prepare next packet of multipacket request

}


void stall_EP0(void)   // ***************************************************
// **************************************************************************
{
volatile uint16_t regVal;

	USB->EP0R = USB->EP0R & CLEAR_CTR_RX_MASK;      //clear CTR_RX
	if (USB->EP0R & 0x0080)  USB->EP0R = USB->EP0R & CLEAR_CTR_TX_MASK; 
													//clear CTR_TX
//toggle EP TX "Stalled": //stalled is DTOG2=0, DTOG1=1
       regVal = USB->EP0R; //toggle bits!!
    if ((regVal & USB_EPTX_DTOG1)==0) USB->EP0R = 0x8290; //USB_EPTX_DTOG2=0x0010
	regVal = USB->EP0R;
	if ((regVal & USB_EPTX_DTOG2)==1) USB->EP0R = 0x82A0; //USB_EPTX_DTOG2=0x0020
//toggle EP RX "Stalled":
	regVal = USB->EP0R;
	if ((regVal & USB_EPRX_DTOG1)==0) USB->EP0R = 0x9280; //USB_EPRX_DTOG1=0x1000
	regVal = USB->EP0R;
	if ((regVal & USB_EPRX_DTOG2)==1) USB->EP0R = 0xA280; //USB_EPRX_DTOG2=0x2000
}


void servEP1_In(void)   // **************************************************
// **************************************************************************
{
volatile uint16_t regVal,temp1,temp2;
uint16_t *EP1Ptr;
uint8_t len;

//clear interrupt flag:
    if (USB->EP1R & 0x0080)  USB->EP1R = USB->EP1R & CLEAR_CTR_TX_MASK;

//fill the buffer:
    EP1Ptr = (uint16_t *) (USB_PMAADDR + EP1_TX_OFFSET);
    len = 0;
    while (gUSB_TXTAIL != gUSB_TXHEAD)  {
        temp2 = 0;
        temp1 = (uint16_t)*gUSB_TXTAIL;
        gUSB_TXTAIL++;
        if (gUSB_TXTAIL == gUSB_TXBUFTOP)  gUSB_TXTAIL = gUSB_TXBUF;
        len++;
        if (len == 64) {
            *EP1Ptr = temp1;
            break;
        }
        if (gUSB_TXTAIL != gUSB_TXHEAD)  {
            temp2 = (uint16_t)*gUSB_TXTAIL;
            gUSB_TXTAIL++;
            if (gUSB_TXTAIL == gUSB_TXBUFTOP)  gUSB_TXTAIL = gUSB_TXBUF;
            len++;
        }
        *EP1Ptr = temp1 | temp2<<8;
        EP1Ptr++;
        if (len == 64) break;
    }
    *gUSB_COUNT1_TX = len;

//make EP TX "VALID":
    regVal = USB->EP1R;
    if ((regVal & USB_EPTX_DTOG2)==0) USB->EP1R = 0x80A1; //USB_EPTX_DTOG2=0x0020
    if ((regVal & USB_EPTX_DTOG1)==0) USB->EP1R = 0x8091; //USB_EPTX_DTOG2=0x0010
}


void servEP3_In(void)   // **************************************************
// **************************************************************************
{
volatile uint16_t regVal;

//clear interrupt flag:
    if (USB->EP3R & 0x0080)  USB->EP3R = USB->EP3R & CLEAR_CTR_TX_MASK;
//effectively there is nothing to notify
//fill the buffer = always send ZLP packets
    *gUSB_COUNT3_TX = 0;
//make EP TX "VALID":
    regVal = USB->EP3R;
    if ((regVal & USB_EPTX_DTOG2)==0) USB->EP3R = 0x86A3; //USB_EPTX_DTOG2=0x0020
    regVal = USB->EP3R;
    if ((regVal & USB_EPTX_DTOG1)==0) USB->EP3R = 0x8693; //USB_EPTX_DTOG2=0x0010
}


// ##########################################################################

// Application level USB interface
//********************************

uint32_t pullUsbRxBuf(void) // **********************************************
// **************************************************************************
{
uint32_t regVal, recv, fillstate;

	servSystem();   //serve watchdog, LED etc.

    if (guConfigured == 0)  return 0;

// check first if something received. If, then set Bit15. If not, return 0
    if (gUSB_RXTAIL != gUSB_RXHEAD)
    {
        recv= (uint32_t)(*gUSB_RXTAIL++ & 0x000000FF);
        if (gUSB_RXTAIL == gUSB_RXBUFTOP)  gUSB_RXTAIL = gUSB_RXBUF;

//calculate fill level of gUSB_RXBUF
//TAIL>HEAD? then add USBBUFLEN to get absolute difference
        if (gUSB_RXTAIL > gUSB_RXHEAD) {
            fillstate =(uint32_t)gUSB_RXHEAD+USB_BUFLEN-(uint32_t)gUSB_RXTAIL;
        }
        else fillstate = (uint32_t)gUSB_RXHEAD - (uint32_t)gUSB_RXTAIL ;

// unlock EP2 interrupt when EP2_BUF contains less than (SIZE-64) bytes
        if ((guLockEP2) && (fillstate < USB_BUFLEN - 70))  {
            guLockEP2 = 0;
//make EP RX "VALID" again:
            regVal = USB->EP2R;
            if ((regVal & USB_EPRX_DTOG2)==0) USB->EP2R = 0xA082; 
													//USB_EPRX_DTOG2=0x2000
            if ((regVal & USB_EPRX_DTOG1)==0) USB->EP2R = 0x9082; 
													//USB_EPRX_DTOG2=0x1000
        }
        LED_Off();		//forward LED handling to main.c (avoid HW dependency)
		return (recv | 0x00008000);
		
//removed 11June22:
/*  
#ifdef MINI
//		if (recv == CTRL_B) gfTermCtrlBD = CTRL_B;		
		if (recv != CTRL_D) return (recv | 0x00008000);
		else  {
			op_DEBUG();		//execute immediately
			return 0;		//don't forward CTRL_D
		}
#else
		if ((gSplit != 2) && (gSplit != 4)) {
		//USB is terminal and MiniDmx is not active
		//particularly ignore CTRL_D if USB is serial I/O
//			if (recv == CTRL_B) gfTermCtrlBD = CTRL_B;		
			if (recv != CTRL_D) return (recv | 0x00008000);
			else  {
				op_DEBUG();		//execute immediately
				return 0;		//don't forward CTRL_D
			}
		}
		else  return (recv | 0x00008000);	//ignore CTRL_D in any case
#endif
*/
    }
	else return 0;   // tail = head
}


uint8_t deCodeMidi(uint8_t len)  // *****************************************
//decode MIDISTREAMING packets to USART MIDI byte stream
//may only be called from "servEP2_OUT" interrupt routine!
// ****************************Ȫ*********************************************
{
uint8_t index, status, data1, data2;

//regular MIDI packets always consist of 4 bytes:
    index = *guEP2Ptr++;
    len-- ;
    if (len == 0)  return 0;
    status = *guEP2Ptr++;
    len-- ;
    if (len == 0)  return 0;
    data1 = *guEP2Ptr++;
    len-- ;
    if (len == 0)  return 0;
    data2 = *guEP2Ptr++;
    len-- ;


    if (index > 0x0F)  return len;
        //wrong "cable number", does not correspond wɩth embedded MIDI IN jack
    if (index < 2)  return len; //reserved for future, seems transmission error
    if ((index == 5) || (index == 0x0F))  {
//1 byte System Common, SysEx or undefined content
#ifdef RUNNINGSTATE
        guRx2RunState = 0;			//invalidate running state
#endif
        *gUSB_RXHEAD++ = status ;   //send one byte in any case
        if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
        return len;
    }
    if ((index == 0x0C)  || (index == 0x0D))  {
//2 bytes MIDI Channel Message
#ifdef RUNNINGSTATE
        if (guRx2RunState != status)  {
            *gUSB_RXHEAD++ = status ;
            if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
            guRx2RunState = status;
        }
#else
        *gUSB_RXHEAD++ = status ;
        if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
#endif

        *gUSB_RXHEAD++ = data1 ;
        if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
        return len;
	}
	if ((index == 2) || (index == 6))  {
//2 bytes System Common or SysEX content
        guRx2RunState = 0;
        *gUSB_RXHEAD++ = status ;
        if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
        *gUSB_RXHEAD++ = data1 ;
        if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
        return len;
	}

//all yet unfiltered packets have 3 bytes content:
#ifdef RUNNINGSTATE
    if (index < 8)  {   //don't use running state with Realtime Messages
        guRx2RunState = 0;
        *gUSB_RXHEAD++ = status ;
        if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
    }
    else  {         //Channel Message
        if (guRx2RunState != status)  {
            *gUSB_RXHEAD++ = status ;
            if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
            guRx2RunState = status;
        }
    }
#else
    *gUSB_RXHEAD++ = status ;
    if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
#endif

    *gUSB_RXHEAD++ = data1 ;
    if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
    *gUSB_RXHEAD++ = data2 ;
    if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
    return len;
}


void sendUsbByte(uint8_t byte)  // ******************************************
// add data byte to EP1 transmit buffer
// **************************************************************************
{
uint32_t fillstate;

    if (guConfigured == 0)  return;
	if (guDeviceClass == CDC_CLASS)  {
		*gUSB_TXHEAD++ = byte;
		if (gUSB_TXHEAD == gUSB_TXBUFTOP)  gUSB_TXHEAD=gUSB_TXBUF;
//added 01Aug2020
//calculate fill level of gUSB_TXBUF
//TAIL>HEAD? then add TXBUFLEN to get the absolute difference
		do {
			servSystem();		//serve not  time critial low level routines
			if (gUSB_TXTAIL > gUSB_TXHEAD) {
			fillstate=(uint32_t)gUSB_TXHEAD +USB_BUFLEN-(uint32_t)gUSB_TXTAIL;
			}
			else fillstate = (uint32_t)gUSB_TXHEAD - (uint32_t)gUSB_TXTAIL ;
//keep reserve of at least 2 USB data packets:
		} while (fillstate > (USB_BUFLEN - 140));//buffer emptied by interrupt
	}
	else  enCodeMidi(byte);
}


void enCodeMidi(uint8_t recv)   // ******************************************
//prepare USART MIDI byte stream for transmission as USB MIDISTREAMING packets
//take one byte per loop
// **************************************************************************
{
uint8_t msgType;
volatile uint8_t *tempHead;

    guTxPackCnt++ ;

    if (recv & 0x80)   {   //recv is status byte
        if (recv != EOX)  guTxPackCnt = 1;  //chance to resync corrupted byte stream
        if (recv == 0xF0)   {
//handle beginning of SysEx message
//special case: no running state, but has a data body which looks like running state
            guMSTATUS = 0xF0;
            guMSTATUSok = 0xFF;
            guMDATA1ok = 0;
            guMDATA2ok = 0;
            guSYSEXBODY = 0;
            return;			//wait for subsequent bytes
        }
//collect 1-3 bytes before packet is sent
        if (recv == EOX)  {			//end of SysEx
            if (guTxPackCnt == 1)  {
//PACKTYPE=0x05: send a 1 byte payload (=EOX) packet
                guPACKTYPE = 5;
                guMSTATUS = EOX;
                guMIDIDAT1 = 0;
                guMIDIDAT2 = 0;
                guMSTATUSok = 0;  //no running state with 1 byte messages
                guMDATA1ok = 0;
                guMDATA2ok = 0;

                if (guSYSEXBODY)   {
                    guMSTATUS = EOX;
                    guMIDIDAT1 = 0;
                    guMIDIDAT2 = 0;
                    guSYSEXBODY = 0;
                    goto SEND_AND_CLEANUP;
                }
                else   {
                    guMSTATUSok = 0;        //undefined MIDI stream
                    guTxPackCnt = 0;        //inore, no error message
                    guMDATA1ok = 0;
                    guMDATA2ok = 0;
                    guSYSEXBODY = 0;
                    return;
                }
            }
            else if (guTxPackCnt == 2)  {   //EOX is last of 2 byte packet
//PACKTYPE=0x06:SysEx ends with 2 bytes. Send 2 byte payload packet
                guPACKTYPE = 6;
                if (guMSTATUS != 0xF0) guMSTATUS = guMIDIDAT1;
                                //if (guMSTATUS==0xF0) incomplete SYSEX Msg
                guMSTATUS = guMIDIDAT1;   //else reorganize
                guMIDIDAT1 = EOX;
                guMIDIDAT2= 0;
                guMSTATUSok = 0;  //no running state with SysEx
                guMDATA1ok = 0;
                guMDATA2ok = 0;
                guSYSEXBODY = 0;

                goto SEND_AND_CLEANUP;
            }
            else   {     //EOX is last of 3 byte packet
//PACKTYPE=0x07: SysEx ends with 3 bytes. Send 3 byte payload packet
                guPACKTYPE = 7;
                if ((guMSTATUS==0xF0) && (guSYSEXBODY==0)) guMIDIDAT2 = EOX;
                                                            //tiny SysEx Msg
                else   {
                    guMSTATUS = guMIDIDAT1;   //reorganize
                    guMIDIDAT1 = guMIDIDAT2;
                    guMIDIDAT2 = EOX;
                }
                guMSTATUSok = 0;  //no running state with SysEx
                guMDATA1ok = 0;
                guMDATA2ok = 0;
                guSYSEXBODY = 0;

                goto SEND_AND_CLEANUP;
            }
        }
//handle different status bytes
        guMSTATUS = recv;
        if (guMSTATUS < 0xF4) { //1 or 2 data bytes Channel- or System Common Msg
	        guMSTATUSok = 0xFF;
            guMDATA1ok = 0;
            guMDATA2ok = 0;
            guSYSEXBODY = 0;
            return;             //wait for subsequent data byte(s)
        }
        if ((guMSTATUS==0xF4)||(guMSTATUS==0xF5)||(guMSTATUS==0xFD)) {
                                            //these message types are undefined
 	        guMSTATUSok = 0;
            guMDATA1ok = 0;
            guMDATA2ok = 0;
            guSYSEXBODY = 0;
            return;
        }
        else  {
//handle remaining 1 byte messages:
//0xF6 = Tune Request
//0xF8 = Timing Clock
//0xFA = MIDI Start
//0xFB = MIDI Continue
//0xFC = MIDI Stop
//0xFE = Active Sensing
//0xFF = MIDI Reset
//PACKTYPE=0x05: send a 1 byte payload packet
            guPACKTYPE = 5;
            guMIDIDAT1 = 0;
            guMIDIDAT2 = 0;
 	        guMSTATUSok = 0;
            guMDATA1ok = 0;
            guMDATA2ok = 0;
            guSYSEXBODY = 0;

            goto SEND_AND_CLEANUP;
        }
    }
    else  {   //handle data bytes (< 0x80)
        msgType = guMSTATUS & 0xF0;
        if (guMSTATUSok == 0)  {    //undefined MIDI stream
            guTxPackCnt = 0;        //inore, no error message
            guMDATA1ok = 0;
            guMDATA2ok = 0;
            guSYSEXBODY = 0;
            return;
        }
        if (guMDATA1ok == 0)   {    //first data byte
            guMIDIDAT1 = recv;      //keep data byte 1 in mind
            guMIDIDAT2 = 0;		    //preemptive for messages with 1 data byte

            if ((guMSTATUS == 0xF1) || (guMSTATUS == 0xF3))  	{
                //2 byte message Timecode or 2 byte message Song Select ready
                //PACKTYPE=0x02: send 2 byte payload packet
                guPACKTYPE = 2;
                guMSTATUSok = 0;    //no Running State with System Common
                guMDATA1ok = 0;
                guMDATA2ok = 0;
                guSYSEXBODY = 0;

             goto SEND_AND_CLEANUP;
            }
            if ((msgType == 0xC0) || (msgType == 0xD0))  {
                //PACKTYPE=0x0C or 0x0D: Channel Msg: send 2 byte payload packet
                guPACKTYPE = (msgType>>4) & 0x0F;
                guMSTATUSok	= 0xFF;     //confirm Running State for next packet
                guMDATA1ok = 0;
                guMDATA2ok = 0;
                guSYSEXBODY = 0;

                goto SEND_AND_CLEANUP;
            }
            else {
                guMDATA1ok = 0xFF;
                return;                 //wait for data byte 2
            }
        }
        else   {  //data byte 2 found
            if (guMSTATUSok == 0) {		//integry check
                guTxPackCnt = 0;        //inore, no error message
                guMDATA1ok = 0;
                guMDATA2ok = 0;
                guSYSEXBODY = 0;
                return;
            }
            if (guMDATA2ok == 0)  {
                guMIDIDAT2 = recv;
                if (guSYSEXBODY)  {
                    guMDATA2ok = 0xFF;
                    return;			    //SysEx body: wait for last byte of packet
                }

                if (guMSTATUS == 0xF2) {  //3 byte message Song Position Pointer ready
                                          // PACKTYPE=0x03: send 3 byte payload packet
                    guPACKTYPE = 3;
                    guMSTATUSok = 0;      //no running state with System Common
                    guMDATA1ok = 0;
                    guMDATA2ok = 0;
                    guSYSEXBODY = 0;

                    goto SEND_AND_CLEANUP;
                }
                else if (guMSTATUS == 0xF0)  { //first 3 bytes of SysEx message ready
                                        //PACKTYPE=0x04: send 3 byte payload packet
                    guPACKTYPE = 4;
                    guSYSEXBODY = 0xFF;
                                    //subsequent data bytes are part of SysEx message
                    guMSTATUSok = 0xFF; //confirm for next SysEx packet
                    guMDATA1ok = 0;
                    guMDATA2ok = 0;

                    goto SEND_AND_CLEANUP;
                }
                else  {
                    //PACKTYPE=0x08,0x09,0x0A,0x0B,0x0E: send 3 byte payload packet
                    guPACKTYPE = (msgType>>4) & 0x0F;
                    guMSTATUSok = 0xFF;  //confirm Running State for next packet
                    guMDATA1ok = 0;
                    guMDATA2ok = 0;

                    goto SEND_AND_CLEANUP;
                }
            }
            else  {                 //special case: 3rd data byte within a SysEx message
                                    //reorganize to get a standard packet
                guMSTATUS = guMIDIDAT1;
                guMIDIDAT1 = guMIDIDAT2;
                guMIDIDAT2 = recv;
                guPACKTYPE = 4;     //PACKTYPE=0x04: send 3 byte payload packet
                guSYSEXBODY = 0xFF; //subsequent data bytes are part of SysEx message
                guMSTATUSok = 0xFF; //confirm for next pack
                guMDATA1ok = 0;
                guMDATA2ok = 0;
            }
        }
    }

SEND_AND_CLEANUP:

    if (guConfigured == 0)  return;
// make sure that the buffer header is always incremented by a complete MIDI packet
// because the USB interrupt is asynchronous with this application level routine
// and blocking of the USB interrupt shall be avoided
    tempHead = gUSB_TXHEAD;
    *tempHead++ = guPACKTYPE;
    if (tempHead == gUSB_TXBUFTOP)  tempHead = gUSB_TXBUF;
     *tempHead++ = guMSTATUS;
    if (tempHead == gUSB_TXBUFTOP)  tempHead = gUSB_TXBUF;
    *tempHead++ = guMIDIDAT1;
    if (tempHead == gUSB_TXBUFTOP)  tempHead = gUSB_TXBUF;
    *tempHead++ = guMIDIDAT2;
    if (tempHead == gUSB_TXBUFTOP)  tempHead = gUSB_TXBUF;
    gUSB_TXHEAD = tempHead;

    guTxPackCnt = 0;      //ready for next package
    guMDATA1ok = 0;
    guMDATA2ok = 0;
    if (((guPACKTYPE<8) && (guPACKTYPE!=4)) || (guMSTATUS > 0xF0)) guMSTATUSok = 0;
            // else guMSTATUSok and guMSTATUS have to be preserved for Running State
            // guSYSEXBODY can only be reset by EOX or non SysEx MIDI message

}


//##########################################################################

